import math

import synapse.exc as s_exc
import synapse.lib.stormtypes as s_stormtypes


# used as a reference implementation:
# https://www.first.org/cvss/calculator/cvsscalc31.js

CVSS31 = {
  'AV': {'N': 0.85, 'A': 0.62, 'L': 0.55, 'P': 0.2},
  'AC': {'H': 0.44, 'L': 0.77},
  'PR': {'U': {'N': 0.85, 'L': 0.62, 'H': 0.27},  # These values are used if Scope is Unchanged
         'C': {'N': 0.85, 'L': 0.68, 'H': 0.5}},  # These values are used if Scope is Changed
  'UI': {'N': 0.85, 'R': 0.62},
  'S': {'U': 6.42, 'C': 7.52},
  'C': {'N': 0, 'L': 0.22, 'H': 0.56},
  'I': {'N': 0, 'L': 0.22, 'H': 0.56},
  'A': {'N': 0, 'L': 0.22, 'H': 0.56},
  'E': {'X': 1, 'U': 0.91, 'P': 0.94, 'F': 0.97, 'H': 1},
  'RL': {'X': 1, 'O': 0.95, 'T': 0.96, 'W': 0.97, 'U': 1},
  'RC': {'X': 1, 'U': 0.92, 'R': 0.96, 'C': 1},
  'CR': {'X': 1, 'L': 0.5, 'M': 1, 'H': 1.5},
  'IR': {'X': 1, 'L': 0.5, 'M': 1, 'H': 1.5},
  'AR': {'X': 1, 'L': 0.5, 'M': 1, 'H': 1.5},
}

vect2prop = {
    'AV': 'cvss:av',
    'AC': 'cvss:ac',
    'PR': 'cvss:pr',
    'UI': 'cvss:ui',
    'S': 'cvss:s',
    'C': 'cvss:c',
    'I': 'cvss:i',
    'A': 'cvss:a',
    'E': 'cvss:e',
    'RL': 'cvss:rl',
    'RC': 'cvss:rc',
    'CR': 'cvss:cr',
    'IR': 'cvss:ir',
    'AR': 'cvss:ar',
    'MS': 'cvss:ms',
    'MC': 'cvss:mc',
    'MI': 'cvss:mi',
    'MA': 'cvss:ma',
    'MAV': 'cvss:mav',
    'MAC': 'cvss:mac',
    'MPR': 'cvss:mpr',
    'MUI': 'cvss:mui',
}

def roundup(x):
    return (math.ceil(x * 10) / 10.0)

@s_stormtypes.registry.registerLib
class CvssLib(s_stormtypes.Lib):
    '''
    A Storm library which implements CVSS score calculations.
    '''
    _storm_locals = (
        {'name': 'calculate', 'desc': 'Calculate the CVSS score values for an input risk:vuln node.',
         'type': {'type': 'function', '_funcname': 'calculate',
                  'args': (
                      {'name': 'node', 'type': 'storm:node',
                       'desc': 'A risk:vuln node from the Storm runtime.'},
                      {'name': 'save', 'type': 'boolean', 'default': True,
                       'desc': 'If true, save the computed scores to the node properties.'},
                      {'name': 'vers', 'type': 'str', 'default': '3.1',
                       'desc': 'The version of CVSS calculations to execute.'},
                  ),
                  'returns': {'type': 'dict', 'desc': 'A dictionary containing the computed score and subscores.', }
        }},
        {'name': 'calculateFromProps', 'desc': 'Calculate the CVSS score values from a props dict.',
         'type': {'type': 'function', '_funcname': 'calculateFromProps',
                  'args': (
                      {'name': 'props', 'type': 'dict',
                       'desc': 'A props dictionary.'},
                      {'name': 'vers', 'type': 'str', 'default': '3.1',
                       'desc': 'The version of CVSS calculations to execute.'},
                  ),
                  'returns': {'type': 'dict', 'desc': 'A dictionary containing the computed score and subscores.', }
        }},
        {'name': 'vectToProps', 'desc': 'Parse a CVSS vector and return a dictionary of risk:vuln props.',
         'type': {'type': 'function', '_funcname': 'vectToProps',
                  'args': (
                      {'name': 'text', 'type': 'str', 'desc': 'A CVSS vector string.'},
                  ),
                  'returns': {'type': 'dict', 'desc': 'A dictionary of risk:vuln secondary props.', }
        }},
        {'name': 'saveVectToNode', 'desc': 'Parse a CVSS vector and record properties on a risk:vuln node.',
         'type': {'type': 'function', '_funcname': 'saveVectToNode',
                  'args': (
                      {'name': 'node', 'type': 'storm:node',
                       'desc': 'A risk:vuln node to record the CVSS properties on.'},
                      {'name': 'text', 'type': 'str', 'desc': 'A CVSS vector string.'},
                  ),
                  'returns': {'type': 'null', }
        }},
    )
    _storm_lib_path = ('infosec', 'cvss',)

    def getObjLocals(self):
        return {
            'calculate': self.calculate,
            'vectToProps': self.vectToProps,
            'saveVectToNode': self.saveVectToNode,
            'calculateFromProps': self.calculateFromProps,
        }

    async def vectToProps(self, text):
        text = await s_stormtypes.tostr(text)
        props = {}

        try:
            for item in text.split('/'):

                name, valu = item.split(':')

                prop = vect2prop.get(name)
                if prop is None:
                    mesg = f'Invalid Vector Element: {name}'
                    raise s_exc.BadArg(mesg=mesg)

                props[prop] = valu

        except ValueError:
            mesg = f'Invalid CVSS Vector: {text}'
            raise s_exc.BadArg(mesg=mesg) from None

        return props

    async def saveVectToNode(self, node, text):
        props = await self.vectToProps(text)
        for prop, valu in props.items():
            await node.set(prop, valu)

    async def calculate(self, node, save=True, vers='3.1'):

        save = await s_stormtypes.tobool(save)

        if node.ndef[0] != 'risk:vuln':
            mesg = '$lib.infosec.cvss.calculate() requires a risk:vuln node.'
            raise s_exc.BadArg(mesg=mesg)

        rval = await self.calculateFromProps(node.props, vers=vers)

        if save and rval.get('ok'):

            score = rval.get('score')
            if score is not None:
                await node.set('cvss:score', score)

            scores = rval.get('scores', {})

            basescore = scores.get('base')
            if basescore is not None:
                await node.set('cvss:score:base', basescore)

            temporalscore = scores.get('temporal')
            if temporalscore is not None:
                await node.set('cvss:score:temporal', temporalscore)

            environmentalscore = scores.get('environmental')
            if environmentalscore is not None:
                await node.set('cvss:score:environmental', environmentalscore)

        return rval

    async def calculateFromProps(self, props, vers='3.1'):

        vers = await s_stormtypes.tostr(vers)

        if vers != '3.1':
            raise s_exc.BadArg(mesg='Currently only vers=3.1 is supported.')

        AV = props.get('cvss:av')
        AC = props.get('cvss:ac')
        PR = props.get('cvss:pr')
        UI = props.get('cvss:ui')
        S = props.get('cvss:s')
        C = props.get('cvss:c')
        _I = props.get('cvss:i')
        A = props.get('cvss:a')

        score = None
        impact = None
        basescore = None
        modimpact = None
        temporalscore = None
        exploitability = None
        environmentalscore = None

        if all((AV, AC, PR, UI, S, C, _I, A)):

            iscbase = 1.0 - ((1.0 - CVSS31['C'][C]) * (1.0 - CVSS31['I'][_I]) * (1.0 - CVSS31['A'][A]))

            if S == 'U':
                impact = 6.42 * iscbase
            else:
                impact = 7.52 * (iscbase - 0.029) - 3.25 * (iscbase - 0.02) ** 15

            exploitability = 8.22 * CVSS31['AV'][AV] * CVSS31['AC'][AC] * CVSS31['PR'][S][PR] * CVSS31['UI'][UI]

            if impact <= 0:
                basescore = 0
            elif S == 'U':
                basescore = min(roundup(impact + exploitability), 10.0)
            else:
                basescore = min(roundup(1.08 * (impact + exploitability)), 10.0)

            score = basescore

            if basescore:

                E = props.get('cvss:e')
                RL = props.get('cvss:rl')
                RC = props.get('cvss:rc')

                if all((E, RL, RC)):
                    expfactor = CVSS31['E'][E] * CVSS31['RL'][RL] * CVSS31['RC'][RC]
                    temporalscore = roundup(basescore * expfactor)
                    score = temporalscore
                else:
                    expfactor = CVSS31['E']['X'] * CVSS31['RL']['X'] * CVSS31['RC']['X']

                MAV = props.get('cvss:mav')
                MAC = props.get('cvss:mac')
                MPR = props.get('cvss:mpr')
                MUI = props.get('cvss:mui')

                MS = props.get('cvss:ms')
                MC = props.get('cvss:mc')
                MI = props.get('cvss:mi')
                MA = props.get('cvss:ma')
                CR = props.get('cvss:cr')
                IR = props.get('cvss:ir')
                AR = props.get('cvss:ar')

                if all((MAV, MAC, MPR, MUI, MS, MC, MI, MA, CR, IR, AR)):

                    if MAV == 'X': MAV = AV
                    if MAC == 'X': MAC = AC
                    if MPR == 'X': MPR = PR
                    if MUI == 'X': MUI = UI
                    if MS == 'X': MS = S
                    if MC == 'X': MC = C
                    if MI == 'X': MI = _I
                    if MA == 'X': MA = A

                    modiscbase = min(1.0 - (
                        (1.0 - (CVSS31['C'][MC] * CVSS31['CR'][CR])) *
                        (1.0 - (CVSS31['I'][MI] * CVSS31['IR'][IR])) *
                        (1.0 - (CVSS31['A'][MA] * CVSS31['AR'][AR]))
                    ), 0.915)

                    mav = CVSS31['AV'].get(MAV, 1)
                    mac = CVSS31['AC'].get(MAC, 1)
                    mpr = CVSS31['PR'][MS].get(MPR, 1)
                    mui = CVSS31['UI'].get(MUI, 1)

                    modexploit = 8.22 * mav * mac * mpr * mui

                    if MS == 'U':
                        modimpact = 6.42 * modiscbase
                    else:
                        modimpact = 7.52 * (modiscbase - 0.029) - 3.25 * (modiscbase * 0.9731 - 0.02) ** 13

                    if modimpact <= 0:
                        environmentalscore = 0
                    elif MS == 'U':
                        environmentalscore = roundup(roundup(min(modimpact + modexploit, 10.0)) * expfactor)
                    else:
                        environmentalscore = roundup(roundup(min(1.08 * (modimpact + modexploit), 10.0)) * expfactor)

                    score = environmentalscore

        rval = {
            'ok': True,
            'version': '3.1',
            'score': score,
            'scores': {
                'base': basescore,
                'temporal': temporalscore,
                'environmental': environmentalscore,
            },
        }

        if impact: rval['scores']['impact'] = round(impact, 1)
        if modimpact: rval['scores']['modifiedimpact'] = round(modimpact, 1)
        if exploitability: rval['scores']['exploitability'] = round(exploitability, 1)

        return rval



import synapse.exc as s_exc
import synapse.tests.utils as s_utils

import synapse.common as s_common

class EconTest(s_utils.SynTest):

    async def test_model_econ(self):

        async with self.getTestCore() as core:

            # test card number 4024007150779444
            card = (await core.nodes('[ econ:pay:card="*" :expr=201802 :name="Bob Smith" :cvv=123 :pin=1234 :pan=4024007150779444 ]'))[0]
            self.eq('bob smith', card.get('name'))
            self.eq(1517443200000, card.get('expr'))
            self.eq('4024007150779444', card.get('pan'))
            self.eq(4, card.get('pan:mii'))
            self.eq(402400, card.get('pan:iin'))

            place = s_common.guid()
            bycont = s_common.guid()
            fromcont = s_common.guid()

            text = f'''[
                econ:purchase="*"

                    :price=13.37
                    :currency=USD
                    :by:contact={bycont}
                    :from:contact={fromcont}

                    :time=20180202
                    :place={place}

                    :paid=true
                    :paid:time=20180202

                    :settled=20180205
            ]'''

            perc = (await core.nodes(text))[0]

            self.eq('13.37', perc.get('price'))
            self.eq('usd', perc.get('currency'))

            self.len(1, await core.nodes('econ:purchase:price=13.37'))
            self.len(1, await core.nodes('econ:purchase:price=13.370'))
            self.len(0, await core.nodes('econ:purchase:price=13.372'))

            with self.raises(s_exc.BadTypeValu):
                await core.nodes('econ:purchase [ :price=730750818665451459101843 ]')

            with self.raises(s_exc.BadTypeValu):
                await core.nodes('econ:purchase [ :price=-730750818665451459101843 ]')

            self.len(1, await core.nodes('econ:purchase:price*range=(13,14)'))

            self.len(1, await core.nodes('econ:purchase:price>10.00'))
            self.len(1, await core.nodes('econ:purchase:price<20.00'))
            self.len(1, await core.nodes('econ:purchase:price>=10.00'))
            self.len(1, await core.nodes('econ:purchase:price>=13.37'))
            self.len(1, await core.nodes('econ:purchase:price<=20.00'))
            self.len(1, await core.nodes('econ:purchase:price<=13.37'))

            self.len(0, await core.nodes('econ:purchase:price<10.00'))
            self.len(0, await core.nodes('econ:purchase:price>20.00'))
            self.len(0, await core.nodes('econ:purchase:price>=20.00'))
            self.len(0, await core.nodes('econ:purchase:price<=10.00'))

            # runtime filter/cmpr test for econ:price
            self.len(1, await core.nodes('econ:purchase:price +:price=13.37'))
            self.len(1, await core.nodes('econ:purchase:price +:price=13.370'))
            self.len(0, await core.nodes('econ:purchase:price +:price=13.372'))

            self.len(1, await core.nodes('econ:purchase:price +:price*range=(13,14)'))

            self.len(1, await core.nodes('econ:purchase:price +:price>10.00'))
            self.len(1, await core.nodes('econ:purchase:price +:price<20.00'))
            self.len(1, await core.nodes('econ:purchase:price +:price>=10.00'))
            self.len(1, await core.nodes('econ:purchase:price +:price>=13.37'))
            self.len(1, await core.nodes('econ:purchase:price +:price<=20.00'))
            self.len(1, await core.nodes('econ:purchase:price +:price<=13.37'))

            self.len(0, await core.nodes('econ:purchase:price +:price<10.00'))
            self.len(0, await core.nodes('econ:purchase:price +:price>20.00'))
            self.len(0, await core.nodes('econ:purchase:price +:price>=20.00'))
            self.len(0, await core.nodes('econ:purchase:price +:price<=10.00'))

            self.eq(bycont, perc.get('by:contact'))
            self.eq(fromcont, perc.get('from:contact'))

            self.eq(True, perc.get('paid'))
            self.eq(1517529600000, perc.get('paid:time'))

            self.eq(1517788800000, perc.get('settled'))

            self.eq(1517529600000, perc.get('time'))
            self.eq(place, perc.get('place'))

            self.len(1, await core.nodes('econ:purchase -> geo:place'))
            self.len(2, await core.nodes('econ:purchase -> ps:contact | uniq'))

            acqu = (await core.nodes(f'[ econ:acquired=({perc.ndef[1]}, (inet:fqdn,vertex.link)) ]'))[0]
            self.eq(perc.ndef[1], acqu.get('purchase'))

            self.len(1, await core.nodes('econ:acquired:item:form=inet:fqdn'))
            self.len(1, await core.nodes('inet:fqdn=vertex.link'))

            self.eq(('inet:fqdn', 'vertex.link'), acqu.get('item'))

            text = f'''[
                econ:acct:payment="*"

                    :to:contact={bycont}
                    :to:coinaddr=(btc, 1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2)

                    :from:contact={fromcont}
                    :from:coinaddr=(btc, 1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2)

                    :from:pay:card={card.ndef[1]}
                    :amount = 20.30
                    :currency = usd

                    :time=20180202
                    :purchase={perc.ndef[1]}
            ]'''
            await core.nodes(text)

            self.len(1, await core.nodes('econ:acct:payment +:time@=(2017,2019) +{-> econ:pay:card +:name="bob smith"}'))

            self.len(1, await core.nodes('econ:acct:payment -> econ:purchase'))
            self.len(1, await core.nodes('econ:acct:payment -> econ:pay:card'))
            self.len(2, await core.nodes('econ:acct:payment -> ps:contact | uniq'))

            nodes = await core.nodes('''
                [ econ:fin:exchange=(us,nasdaq) :name=nasdaq :currency=usd :org=* ]
            ''')
            self.len(1, nodes)
            self.nn(nodes[0].ndef[1])
            self.nn(nodes[0].get('org'))
            self.eq('usd', nodes[0].get('currency'))
            self.eq('nasdaq', nodes[0].get('name'))

            nodes = await core.nodes('''
                [
                    econ:fin:security=(us, nasdaq, tsla)
                        :exchange=(us, nasdaq)
                        :ticker=nasdaq/tsla
                        :type=STOCK
                        :price=9999.00
                        :time=202002
                ]
            ''')

            self.len(1, nodes)
            self.eq('947183947f2e2c7bdc55264c20670f19', nodes[0].ndef[1])
            self.eq('stock', nodes[0].get('type'))
            self.eq('nasdaq/tsla', nodes[0].get('ticker'))
            self.eq('9999', nodes[0].get('price'))
            self.eq(1580515200000, nodes[0].get('time'))

            self.len(1, await core.nodes('econ:fin:security -> econ:fin:exchange +:name=nasdaq'))

            nodes = await core.nodes('''
                [
                    econ:fin:tick=*
                        :time=20200202
                        :security=(us, nasdaq, tsla)
                        :price=9999.00
                ]
            ''')
            self.len(1, nodes)
            self.eq(1580601600000, nodes[0].get('time'))
            self.eq('947183947f2e2c7bdc55264c20670f19', nodes[0].get('security'))
            self.eq('9999', nodes[0].get('price'))

            nodes = await core.nodes('''
                [
                    econ:fin:bar=*
                        :ival=(20200202, 20200203)
                        :security=(us, nasdaq, tsla)
                        :price:open=9999.00
                        :price:close=9999.01
                        :price:high=999999999999.00
                        :price:low=0.00001
                ]
            ''')
            self.len(1, nodes)
            self.eq((1580601600000, 1580688000000), nodes[0].get('ival'))
            self.eq('947183947f2e2c7bdc55264c20670f19', nodes[0].get('security'))
            self.eq('9999', nodes[0].get('price:open'))
            self.eq('9999.01', nodes[0].get('price:close'))
            self.eq('999999999999', nodes[0].get('price:high'))
            self.eq('0.00001', nodes[0].get('price:low'))

            nodes = await core.nodes('[ econ:acct:payment=* :from:contract=* :to:contract=* :memo="2012 Chateauneuf du Pape" ]')
            self.len(1, nodes)
            self.eq('2012 Chateauneuf du Pape', nodes[0].get('memo'))
            self.nn(nodes[0].get('to:contract'))
            self.nn(nodes[0].get('from:contract'))
            nodes = await core.nodes('econ:acct:payment :to:contract -> ou:contract')
            self.len(1, nodes)
            nodes = await core.nodes('econ:acct:payment :from:contract -> ou:contract')
            self.len(1, nodes)

            nodes = await core.nodes('''
                [ econ:acct:balance=*
                    :time = 20211031
                    :pay:card = *
                    :crypto:address = btc/12345
                    :amount = 123.45
                    :currency = usd
                    :delta = 12.00
                ]''')
            self.len(1, nodes)
            self.nn(nodes[0].get('pay:card'))
            self.eq(nodes[0].get('time'), 1635638400000)
            self.eq(nodes[0].get('crypto:address'), ('btc', '12345'))
            self.eq(nodes[0].get('amount'), '123.45')
            self.eq(nodes[0].get('currency'), 'usd')
            self.eq(nodes[0].get('delta'), '12')



import asyncio

import synapse.exc as s_exc

import synapse.lib.cli as s_cli
import synapse.lib.version as s_version

import synapse.tests.utils as s_t_utils

class TstThrowCmd(s_cli.Cmd):
    '''
    Throw an exception.
    '''
    _cmd_name = 'throwzero'

    async def runCmdOpts(self, opts):
        1 / 0

class TstThrowKeyboard(s_cli.Cmd):
    '''
    Throw an exception.
    '''
    _cmd_name = 'throwkeyboard'

    async def runCmdOpts(self, opts):
        raise asyncio.CancelledError()


class CliTest(s_t_utils.SynTest):

    async def test_cli_prompt(self):
        outp = self.getTestOutp()
        async with await s_cli.Cli.anit(None, outp=outp) as cli:
            self.eq(cli.getCmdPrompt(), 'cli> ')
            cli.cmdprompt = 'hehe> '
            self.eq(cli.getCmdPrompt(), 'hehe> ')

    async def test_cli_get_set(self):
        outp = self.getTestOutp()
        async with await s_cli.Cli.anit(None, outp=outp, hehe='haha') as cli:
            self.eq(cli.get('hehe'), 'haha')
            self.none(cli.get('foo'))
            cli.set('foo', 'bar')
            self.eq(cli.get('foo'), 'bar')
            await cli.runCmdLine('locs')
            self.true(outp.expect('hehe'))
            self.true(outp.expect('haha'))
            self.true(outp.expect('foo'))
            self.true(outp.expect('bar'))

        outp = self.getTestOutp()
        async with self.getTestCoreAndProxy() as (core, proxy):
            async with await s_cli.Cli.anit(proxy, outp=outp) as cli:
                cli.echoline = True
                await cli.runCmdLine('locs')
                self.true(outp.expect('syn:local:version'))
                self.true(outp.expect('syn:remote:version'))
                self.true(outp.expect(s_version.verstring))

    async def test_cli_quit(self):
        outp = self.getTestOutp()
        async with await s_cli.Cli.anit(None, outp=outp) as cli:
            await cli.runCmdLine('quit')
            self.true(cli.isfini)

    async def test_cli_help(self):
        outp = self.getTestOutp()
        async with await s_cli.Cli.anit(None, outp=outp) as cli:
            await cli.runCmdLine('help')
        self.true(outp.expect('Quit the current command line interpreter.'))

        outp = self.getTestOutp()
        async with await s_cli.Cli.anit(None, outp=outp) as cli:
            await cli.runCmdLine('help newp')
        self.true(outp.expect('=== NOT FOUND: newp'))

    async def test_cli_notacommand(self):
        outp = self.getTestOutp()
        async with await s_cli.Cli.anit(None, outp=outp) as cli:
            await cli.runCmdLine('notacommand')
        self.true(outp.expect('cmd not found: notacommand'))

    async def test_cli_cmdret(self):

        class WootCmd(s_cli.Cmd):
            _cmd_name = 'woot'

            async def runCmdOpts(self, opts):
                return 20

        async with await s_cli.Cli.anit(None) as cli:
            cli.addCmdClass(WootCmd)
            self.eq(await cli.runCmdLine('woot'), 20)

    async def test_cli_cmd(self):
        async with await s_cli.Cli.anit(None) as cli:
            quit = cli.getCmdByName('quit')
            self.nn(quit.getCmdDoc())
            self.nn(quit.getCmdBrief())

    async def test_cli_opts_flag(self):

        async with await s_cli.Cli.anit(None) as cli:

            quit = cli.getCmdByName('quit')

            quit._cmd_syntax = (
                ('--bar', {}),
                ('haha', {'type': 'valu'}),
            )

            opts = quit.getCmdOpts('quit --bar hoho')

            self.eq(opts.get('bar'), True)
            self.eq(opts.get('haha'), 'hoho')

    async def test_cli_opts_list(self):

        async with await s_cli.Cli.anit(None) as cli:

            quit = cli.getCmdByName('quit')

            quit._cmd_syntax = (
                ('--bar', {}),
                ('haha', {'type': 'list'}),
            )

            opts = quit.getCmdOpts('quit --bar hoho haha "hehe hehe"')

            self.eq(opts.get('bar'), True)
            self.eq(tuple(opts.get('haha')), ('hoho', 'haha', 'hehe hehe'))

    async def test_cli_opts_enum(self):
        async with await s_cli.Cli.anit(None) as cli:

            quit = cli.getCmdByName('quit')

            quit._cmd_syntax = (
                ('--enum', {'type': 'enum', 'enum:vals': ('foo', 'bar', 'baz'), 'defval': 'def'}),
            )
            opts = quit.getCmdOpts('quit')
            self.eq(opts.get('enum'), 'def')

            opts = quit.getCmdOpts('quit --enum foo')
            self.eq(opts.get('enum'), 'foo')

            opts = quit.getCmdOpts('quit --enum bar')
            self.eq(opts.get('enum'), 'bar')

            opts = quit.getCmdOpts('quit --enum baz')
            self.eq(opts.get('enum'), 'baz')

            self.raises(s_exc.BadSyntax, quit.getCmdOpts, 'quit --enum newp')

    async def test_cli_opts_glob(self):

        async with await s_cli.Cli.anit(None) as cli:

            quit = cli.getCmdByName('quit')

            quit._cmd_syntax = (
                ('--bar', {}),
                ('haha', {'type': 'glob'}),
            )

            opts = quit.getCmdOpts('quit --bar hoho lulz')

            self.eq(opts.get('bar'), True)
            self.eq(opts.get('haha'), 'hoho lulz')

    async def test_cli_opts_defval(self):

        async with await s_cli.Cli.anit(None) as cli:

            quit = cli.getCmdByName('quit')

            quit._cmd_syntax = (
                ('--bar', {'type': 'valu', 'defval': 'lol'}),
                ('haha', {'type': 'glob'}),
            )

            opts = quit.getCmdOpts('quit hoho lulz')
            self.eq(opts.get('bar'), 'lol')

    async def test_cli_opts_parse_valu(self):

        async with await s_cli.Cli.anit(None) as cli:

            quit = cli.getCmdByName('quit')

            quit._cmd_syntax = (
                ('--bar', {'type': 'valu'}),
            )

            opts = quit.getCmdOpts('quit --bar woah')
            self.eq(opts.get('bar'), 'woah')

            self.raises(s_exc.BadSyntax, quit.getCmdOpts, 'quit --bar woah this is too much text')

    async def test_cli_opts_parse_list(self):

        async with await s_cli.Cli.anit(None) as cli:

            quit = cli.getCmdByName('quit')

            quit._cmd_syntax = (
                ('--bar', {'type': 'list'}),
            )

            # The list must be quoted
            opts = quit.getCmdOpts('quit --bar "1,2,3"')
            self.eq(opts.get('bar'), ['1', '2', '3'])

            # Or encapsulated in a storm list syntax
            opts = quit.getCmdOpts('quit --bar (1, 2, 3)')
            self.eq(opts.get('bar'), [1, 2, 3])

            # A single item is fine
            opts = quit.getCmdOpts('quit --bar woah')
            self.eq(opts.get('bar'), ['woah'])

    async def test_cli_cmd_loop_quit(self):
        outp = self.getTestOutp()
        cmdg = s_t_utils.CmdGenerator(['help', 'quit'])

        async with await s_cli.Cli.anit(None, outp) as cli:
            cli.prompt = cmdg
            await cli.runCmdLoop()
            self.eq(cli.isfini, True)

        self.true(outp.expect('o/'))

    async def test_cli_cmd_loop_eof(self):
        outp = self.getTestOutp()
        cmdg = s_t_utils.CmdGenerator(['help', EOFError()])
        async with await s_cli.Cli.anit(None, outp) as cli:
            cli.prompt = cmdg
            await cli.runCmdLoop()
            self.eq(cli.isfini, True)
        self.false(outp.expect('o/', throw=False))

    async def test_cli_cmd_loop_bad_input(self):
        outp = self.getTestOutp()
        cmdg = s_t_utils.CmdGenerator([1234, EOFError()])
        async with await s_cli.Cli.anit(None, outp) as cli:
            cli.prompt = cmdg
            await cli.runCmdLoop()
            self.eq(cli.isfini, True)
        self.true(outp.expect("AttributeError: 'int' object has no attribute 'strip'", throw=False))

    async def test_cli_cmd_loop_keyint(self):

        outp = self.getTestOutp()

        cmdg = s_t_utils.CmdGenerator(['help', KeyboardInterrupt(), 'quit'])

        async with await s_cli.Cli.anit(None, outp) as cli:
            cli.prompt = cmdg
            await cli.runCmdLoop()
            self.eq(cli.isfini, True)

        self.true(outp.expect('<ctrl-c>'))

    async def test_cli_cmd_loop(self):

        outp = self.getTestOutp()
        cmdg = s_t_utils.CmdGenerator(['help', 'locs', '', '    ', 'throwzero', 'throwkeyboard', 'quit'])

        async with await s_cli.Cli.anit(None, outp) as cli:

            cli.prompt = cmdg

            cli.addCmdClass(TstThrowCmd)
            cli.addCmdClass(TstThrowKeyboard)

            await cli.runCmdLoop()

            self.true(outp.expect('o/'))
            self.true(outp.expect('"syn:local:version"'))
            self.true(outp.expect(f'"{s_version.verstring}"'))
            self.true(outp.expect('ZeroDivisionError'))
            self.true(outp.expect('Cmd cancelled'))
            self.true(cli.isfini)

    async def test_cli_fini_disconnect(self):

        outp = self.getTestOutp()

        async with self.getTestCoreAndProxy() as (core, prox):
            cli = await s_cli.Cli.anit(prox, outp=outp)

        self.true(prox.isfini)
        self.true(core.isfini)
        self.true(cli.isfini)
        self.true(outp.expect('connection closed...'))



import os

import synapse.tools.json2mpk as s_json2mpk

import synapse.tests.utils as s_t_utils

class Json2MpkTest(s_t_utils.SynTest):

    def test_tools_json2mpk(self):

        with self.getTestDir() as ndir:

            outp = self.getTestOutp()

            path = os.path.join(ndir, 'woot.json')
            newp = os.path.join(ndir, 'woot.mpk')
            fake = os.path.join(ndir, 'fake.json')
            html = os.path.join(ndir, 'woot.html')

            self.false(os.path.isfile(path))
            self.false(os.path.isfile(newp))

            with open(path, 'wb') as fd:
                fd.write(b'{"foo":10}\n{"bar":20}\n')

            args = ['--rm', path, fake, html]

            s_json2mpk.main(args, outp=outp)

            with open(newp, 'rb') as fd:
                self.eq(fd.read(), b'\x81\xa3foo\n\x81\xa3bar\x14')

            self.false(os.path.isfile(path))



'''
Time related utilities for synapse "epoch millis" time values.
'''
import datetime

from dateutil.relativedelta import relativedelta

import pytz
import regex

import synapse.exc as s_exc

EPOCH = datetime.datetime(1970, 1, 1)

tz_hm_re = regex.compile(r'\d((\+|\-)(\d{1,2}):(\d{2}))($|(\-\w+|\+\w))')

def _rawparse(text, base=None, chop=False):

    text = text.strip().lower().replace(' ', '')

    if base is None:
        text, base = parsetz(text)

    text = (''.join([c for c in text if c.isdigit()]))

    if chop:
        text = text[:17]

    tlen = len(text)

    try:
        if tlen == 4:
            dt = datetime.datetime.strptime(text, '%Y')

        elif tlen == 6:
            dt = datetime.datetime.strptime(text, '%Y%m')

        elif tlen == 8:
            dt = datetime.datetime.strptime(text, '%Y%m%d')

        elif tlen == 10:
            dt = datetime.datetime.strptime(text, '%Y%m%d%H')

        elif tlen == 12:
            dt = datetime.datetime.strptime(text, '%Y%m%d%H%M')

        elif tlen == 14:
            dt = datetime.datetime.strptime(text, '%Y%m%d%H%M%S')

        elif 15 <= tlen <= 20:
            dt = datetime.datetime.strptime(text, '%Y%m%d%H%M%S%f')

        else:
            raise s_exc.BadTypeValu(valu=text, name='time',
                                    mesg='Unknown time format')
    except ValueError as e:
        raise s_exc.BadTypeValu(mesg=str(e))

    return dt, base, tlen

def parse(text, base=None, chop=False):
    '''
    Parse a time string into an epoch millis value.

    Args:
        text (str): Time string to parse
        base (int or None): Milliseconds to offset the time from
        chop (bool): Whether to chop the digit-only string to 17 chars

    Returns:
        int: Epoch milliseconds
    '''
    dtraw, base, tlen = _rawparse(text, base=base, chop=chop)
    return int((dtraw - EPOCH).total_seconds() * 1000 + base)

def wildrange(text):
    '''
    Parse an interval from a wild card time stamp: 2021/10/31*
    '''
    dttick, base, tlen = _rawparse(text)
    if tlen not in (4, 6, 8, 10, 12, 14):
        mesg = 'Time wild card position not supported.'
        raise s_exc.BadTypeValu(mesg=mesg)

    if tlen == 4:
        dttock = dttick + relativedelta(years=1)
    elif tlen == 6:
        dttock = dttick + relativedelta(months=1)
    elif tlen == 8:
        dttock = dttick + relativedelta(days=1)
    elif tlen == 10:
        dttock = dttick + relativedelta(hours=1)
    elif tlen == 12:
        dttock = dttick + relativedelta(minutes=1)
    else:  # tlen = 14
        dttock = dttick + relativedelta(seconds=1)

    tick = int((dttick - EPOCH).total_seconds() * 1000 + base)
    tock = int((dttock - EPOCH).total_seconds() * 1000 + base)
    return (tick, tock)

def parsetz(text):
    '''
    Parse timezone from time string, with UTC as the default.

    Args:
        text (str): Time string

    Returns:
        tuple: A tuple of text with tz chars removed and base milliseconds to offset time.
    '''
    tz_hm = tz_hm_re.search(text)

    if tz_hm is not None:

        tzstr, rel, hrs, mins, _, _ = tz_hm.groups()

        rel = 1 if rel == '-' else -1

        base = rel * (onehour * int(hrs) + onemin * int(mins))

        if abs(base) >= oneday:
            raise s_exc.BadTypeValu(valu=text, name='time', mesg=f'Timezone offset must be between +/- 24 hours')

        return text.replace(tzstr, '', 1), base

    return text, 0

def repr(tick, pack=False):
    '''
    Return a date string for an epoch-millis timestamp.

    Args:
        tick (int): The timestamp in milliseconds since the epoch.

    Returns:
        (str):  A date time string
    '''
    if tick == 0x7fffffffffffffff:
        return '?'

    dt = EPOCH + datetime.timedelta(milliseconds=tick)
    millis = dt.microsecond / 1000
    if pack:
        return '%d%.2d%.2d%.2d%.2d%.2d%.3d' % (dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, millis)
    return '%d/%.2d/%.2d %.2d:%.2d:%.2d.%.3d' % (dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, millis)

def day(tick):
    return (EPOCH + datetime.timedelta(milliseconds=tick)).day

def year(tick):
    return (EPOCH + datetime.timedelta(milliseconds=tick)).year

def month(tick):
    return (EPOCH + datetime.timedelta(milliseconds=tick)).month

def hour(tick):
    return (EPOCH + datetime.timedelta(milliseconds=tick)).hour

def minute(tick):
    return (EPOCH + datetime.timedelta(milliseconds=tick)).minute

def second(tick):
    return (EPOCH + datetime.timedelta(milliseconds=tick)).second

def dayofmonth(tick):
    return (EPOCH + datetime.timedelta(milliseconds=tick)).day - 1

def dayofweek(tick):
    return (EPOCH + datetime.timedelta(milliseconds=tick)).weekday()

def dayofyear(tick):
    return (EPOCH + datetime.timedelta(milliseconds=tick)).timetuple().tm_yday - 1

def ival(*times):

    times = [t for t in times if t is not None]

    minv = min(times)
    maxv = max(times)

    if minv == maxv:
        maxv += 1

    return (minv, maxv)

onesec = 1000
onemin = 60000
onehour = 3600000
oneday = 86400000

timeunits = {
    'sec': onesec,
    'secs': onesec,
    'seconds': onesec,

    'min': onemin,
    'mins': onemin,
    'minute': onemin,
    'minutes': onemin,

    'hour': onehour,
    'hours': onehour,

    'day': oneday,
    'days': oneday,
}

# TODO: use synapse.lib.syntax once it gets cleaned up
def _noms(text, offs, cset):
    begin = offs
    while len(text) > offs and text[offs] in cset:
        offs += 1
    return text[begin:offs], offs

def delta(text):
    '''
    Parse a simple time delta string and return the delta.
    '''
    text = text.strip().lower()

    _, offs = _noms(text, 0, ' \t\r\n')

    sign = '+'
    if text and text[0] in ('+', '-'):
        sign = text[0]
        offs += 1

    _, offs = _noms(text, offs, ' \t\r\n')

    sizetext, offs = _noms(text, offs, '0123456789')

    _, offs = _noms(text, offs, ' \t\r\n')

    unittext = text[offs:]

    size = int(sizetext, 0)

    if sign == '-':
        size = -size

    base = timeunits.get(unittext)
    if base is None:
        mesg = f'unknown time delta units: {unittext}'
        raise s_exc.BadTypeValu(name='time', valu=text, mesg=mesg)

    return size * base

def toUTC(tick, fromzone):
    try:
        tz = pytz.timezone(fromzone)
    except pytz.exceptions.UnknownTimeZoneError as e:
        mesg = f'Unknown timezone: {fromzone}'
        raise s_exc.BadArg(mesg=mesg) from e

    base = datetime.datetime(1970, 1, 1, tzinfo=tz) + datetime.timedelta(milliseconds=tick)
    return int(base.astimezone(pytz.UTC).timestamp() * 1000)



'''
An RMI framework for synapse.
'''

import os
import time
import asyncio
import logging
import collections

import aiohttp

import synapse.exc as s_exc
import synapse.glob as s_glob
import synapse.common as s_common
import synapse.lib.base as s_base
import synapse.lib.coro as s_coro
import synapse.lib.link as s_link
import synapse.lib.queue as s_queue
import synapse.lib.certdir as s_certdir
import synapse.lib.threads as s_threads
import synapse.lib.urlhelp as s_urlhelp
import synapse.lib.hashitem as s_hashitem

logger = logging.getLogger(__name__)

televers = (3, 0)

aha_clients = {}

async def addAhaUrl(url):
    '''
    Add (incref) an aha registry URL.

    NOTE: You may also add a list of redundant URLs.
    '''
    hkey = s_hashitem.normitem(url)

    info = aha_clients.get(hkey)
    if info is None:
        info = aha_clients[hkey] = {'refs': 0, 'client': None, 'url': url}

    info['refs'] += 1
    return info

async def delAhaUrl(url):
    '''
    Remove (decref) an aha registry URL.

    NOTE: You may also remove a list of redundant URLs.
    '''
    hkey = s_hashitem.normitem(url)

    info = aha_clients.get(hkey)
    if info is None:
        return 0

    info['refs'] -= 1

    refs = info['refs']

    if refs == 0:
        client = info.get('client')
        if client is not None:
            await client.fini()

        aha_clients.pop(hkey, None)

    return refs

def zipurl(info):
    '''
    Reconstruct a URL string from a parsed telepath info dict.
    '''
    # copy to prevent mutation
    info = dict(info)

    host = info.pop('host', None)
    port = info.pop('port', None)
    path = info.pop('path', None)
    user = info.pop('user', None)
    passwd = info.pop('passwd', None)
    scheme = info.pop('scheme', None)

    url = f'{scheme}://'
    if user:
        url += user
        if passwd:
            url += f':{passwd}'
        url += '@'

    if host:
        url += host
        if port is not None:
            url += f':{port}'

    if path:
        url += f'{path}'

    if info:
        params = '&'.join([f'{k}={v}' for (k, v) in info.items()])
        url += f'?{params}'

    return url

def mergeAhaInfo(info0, info1):

    # copy both to prevent mutation
    info0 = dict(info0)
    info1 = dict(info1)

    # local path wins
    info1.pop('path', None)

    # upstream wins everything else
    info0.update(info1)

    return info0

async def getAhaProxy(urlinfo):
    '''
    Return a telepath proxy by looking up a host from an aha registry.
    '''
    host = urlinfo.get('host')
    if host is None:
        mesg = f'getAhaProxy urlinfo has no host: {urlinfo}'
        raise s_exc.NoSuchName(mesg=mesg)

    laste = None
    for ahaurl, cnfo in list(aha_clients.items()):
        client = cnfo.get('client')
        if client is None:
            client = await Client.anit(cnfo.get('url'))
            client._fini_atexit = True
            cnfo['client'] = client

        try:
            proxy = await client.proxy(timeout=5)

            ahasvc = await asyncio.wait_for(proxy.getAhaSvc(host), timeout=5)
            if ahasvc is None:
                continue

            svcinfo = ahasvc.get('svcinfo', {})
            if not svcinfo.get('online'):
                continue

            info = mergeAhaInfo(urlinfo, svcinfo.get('urlinfo', {}))

            return await openinfo(info)

        except asyncio.CancelledError:  # pragma: no cover
            raise

        except Exception as e:
            if isinstance(ahaurl, str):
                surl = s_urlhelp.sanitizeUrl(ahaurl)
            else:
                surl = tuple([s_urlhelp.sanitizeUrl(u) for u in ahaurl])
            logger.exception(f'Unable to get aha client ({surl})')

            laste = e

    if laste is not None:
        raise laste

    mesg = f'aha lookup failed: {host}'
    raise s_exc.NoSuchName(mesg=mesg)

async def loadTeleEnv(path):

    if not os.path.isfile(path):
        return

    conf = s_common.yamlload(path)

    vers = conf.get('version')
    if vers != 1:
        logger.warning(f'telepath.yaml unknown version: {vers}')
        return

    ahas = conf.get('aha:servers', ())
    cdirs = conf.get('certdirs', ())

    for a in ahas:
        await addAhaUrl(a)

    for p in cdirs:
        s_certdir.addCertPath(p)

    async def fini():
        for a in ahas:
            await delAhaUrl(a)
        for p in cdirs:
            s_certdir.delCertPath(p)

    return fini

class Aware:
    '''
    The telepath.Aware mixin allows shared objects to
    handle individual links managed by the Daemon.
    '''
    async def getTeleApi(self, link, mesg, path):
        '''
        Return a shared object for this link.
        Args:
            link (synapse.lib.link.Link): A network link.
            mesg ((str,dict)): The tele:syn handshake message.

        '''
        return self

    def onTeleShare(self, dmon, name):
        pass

class Task:
    '''
    A telepath Task is used to internally track calls/responses.
    '''
    def __init__(self):
        self.retn = None
        self.iden = s_common.guid()
        self.done = asyncio.Event()

    async def result(self):
        await self.done.wait()
        return self.retn

    def reply(self, retn):
        self.retn = retn
        self.done.set()

class Share(s_base.Base):
    '''
    The telepath client side of a dynamically shared object.
    '''
    async def __anit__(self, proxy, iden, sharinfo=None):
        await s_base.Base.__anit__(self)
        self.iden = iden
        self.proxy = proxy

        if sharinfo is None:
            sharinfo = {}
        self.sharinfo = sharinfo
        self.methinfo = sharinfo.get('meths', {})
        self.proxy.shares[iden] = self

        self.txfini = True
        self.onfini(self._txShareFini)

    async def _txShareFini(self):

        self.proxy.shares.pop(self.iden, None)

        if not self.txfini:
            return

        mesg = ('share:fini', {'share': self.iden})
        if not self.proxy.link.isfini:
            await self.proxy.link.tx(mesg)

    def __getattr__(self, name):
        info = self.methinfo.get(name)
        if info is not None and info.get('genr'):
            meth = GenrMethod(self.proxy, name, share=self.iden)
            setattr(self, name, meth)
            return meth

        meth = Method(self.proxy, name, share=self.iden)
        setattr(self, name, meth)
        return meth

    def __enter__(self):
        '''
        Convenience function to enable using Proxy objects as synchronous context managers.

        Note:
            This should never be used by synapse core code.  This is for sync client code convenience only.
        '''
        if s_threads.iden() == self.tid:
            raise s_exc.SynErr('Use of synchronous context manager in async code')

        self._ctxobj = self.schedCoroSafePend(self.__aenter__())
        return self

    def __exit__(self, *args):
        '''
        This should never be used by synapse core code.  This is for sync client code convenience only.
        '''
        return self.schedCoroSafePend(self._ctxobj.__aexit__(*args))

class Genr(Share):

    async def __anit__(self, proxy, iden):
        await Share.__anit__(self, proxy, iden, sharinfo={})
        self.queue = await s_queue.AQueue.anit()
        self.onfini(self.queue.fini)

    async def _onShareData(self, data):
        self.queue.put(data)

    async def __aiter__(self):

        try:
            while not self.isfini:

                for retn in await self.queue.slice():
                    if retn is None:
                        return

                    yield s_common.result(retn)

            raise s_exc.LinkShutDown(mesg='Remote peer disconnected')

        finally:
            await self.fini()

    def __iter__(self):

        try:
            while not self.isfini:

                for retn in s_glob.sync(self.queue.slice()):

                    if retn is None:
                        return

                    yield s_common.result(retn)

        finally:
            s_glob.sync(self.fini())

sharetypes = {
    'share': Share,
    'genr': Genr,
}

class Method:
    '''
    The telepath Method is used to provide proxy method calls.
    '''
    def __init__(self, proxy, name, share=None):
        self.name = name
        self.share = share
        self.proxy = proxy

        # act as much like a bound method as possible...
        self.__name__ = name
        self.__self__ = proxy

    @s_glob.synchelp
    async def __call__(self, *args, **kwargs):
        todo = (self.name, args, kwargs)
        return await self.proxy.task(todo, name=self.share)

class GenrIter:
    '''
    An object to help delay a telepath call until iteration.
    '''
    def __init__(self, proxy, todo, share):
        self.todo = todo
        self.proxy = proxy
        self.share = share

    async def list(self):
        return [x async for x in self]

    async def __aiter__(self):

        genr = await self.proxy.task(self.todo, name=self.share)
        if genr is None:
            return

        async for item in genr:
            yield item
            await asyncio.sleep(0)

    def __iter__(self):
        genr = s_glob.sync(self.proxy.task(self.todo, name=self.share))
        for item in genr:
            yield item

class GenrMethod(Method):

    def __call__(self, *args, **kwargs):
        todo = (self.name, args, kwargs)
        return GenrIter(self.proxy, todo, self.share)

class Pipeline(s_base.Base):

    async def __anit__(self, proxy, genr, name=None):

        await s_base.Base.__anit__(self)

        self.genr = genr
        self.name = name
        self.proxy = proxy

        self.count = 0

        self.link = await proxy.getPoolLink()
        self.task = self.schedCoro(self._runGenrLoop())
        self.taskexc = None

    async def _runGenrLoop(self):

        try:
            async for todo in self.genr:

                mesg = ('t2:init', {
                        'todo': todo,
                        'name': self.name,
                        'sess': self.proxy.sess})

                await self.link.tx(mesg)
                self.count += 1

        except asyncio.CancelledError:
            raise

        except Exception as e:
            self.taskexc = e
            await self.link.fini()
            raise

    async def __aiter__(self):

        taskdone = False
        while not self.isfini:

            if not taskdone and self.task.done():
                taskdone = True
                self.task.result()

            if taskdone and self.count == 0:
                if not self.link.isfini:
                    await self.proxy._putPoolLink(self.link)
                await self.fini()
                return

            mesg = await self.link.rx()
            if self.taskexc:
                raise self.taskexc

            if mesg is None:
                raise s_exc.LinkShutDown(mesg='Remote peer disconnected')

            if mesg[0] == 't2:fini':
                self.count -= 1
                yield mesg[1].get('retn')
                continue

            logger.warning(f'Pipeline got unhandled message: {mesg!r}.')  # pragma: no cover

class Proxy(s_base.Base):
    '''
    A telepath Proxy is used to call remote APIs on a shared object.

    Example:

        import synapse.telepath as s_telepath

        # open the "foo" object shared in a dmon on localhost:3344

        async def doFooThing():

            proxy = await s_telepath.openurl('tcp://127.0.0.1:3344/foo')

            valu = await proxy.getFooValu(x, y)

    The proxy (and openurl function) may also be used from sync code:

        proxy = s_telepath.openurl('tcp://127.0.0.1:3344/foo')

        valu = proxy.getFooValu(x, y)

    '''
    async def __anit__(self, link, name):

        await s_base.Base.__anit__(self)
        self.tid = s_threads.iden()

        self.link = link
        self.name = name

        self.tasks = {}
        self.shares = {}

        self.sharinfo = {}
        self.methinfo = {}

        self.sess = None
        self.links = collections.deque()
        self._link_poolsize = 4

        self.synack = None
        self.syndone = asyncio.Event()

        self.handlers = {
            'task:fini': self._onTaskFini,
            'share:data': self._onShareData,
            'share:fini': self._onShareFini,
        }

        async def fini():

            for item in list(self.shares.values()):
                await item.fini()

            mesg = ('task:fini', {'retn': (False, ('IsFini', {}))})
            for name, task in list(self.tasks.items()):
                task.reply(mesg)
                del self.tasks[name]

            for link in self.links:
                await link.fini()

            del self.syndone
            await self.link.fini()

        self.onfini(fini)
        self.link.onfini(self.fini)

    def _getSynVers(self):
        '''
        Helper method to retrieve the remote Synapse version from Proxy.

        Notes:
            This will return None if the synapse version was not supplied
            during the Telepath handshake.

        Returns:
            tuple: A tuple of major, minor, patch information as integers.
        '''
        version = self.sharinfo.get('syn:version')
        return version

    def _getSynCommit(self):
        '''
        Helper method to retrieve the remote Synapse commit hash from Proxy.

        Notes:
            This will return None if the synapse commit hash was not supplied
            during the Telepath handshake.

        Returns:
            str: A string containing the commit hash. This may be a empty string.
        '''
        return self.sharinfo.get('syn:commit')

    def _getClasses(self):
        '''
        Helper method to retrieve the classes that comprise the remote object.

        Notes:
            This will return None if the class version was not supplied
            during the Telepath handshake.

        Returns:
            tuple: A tuple of strings containing the class paths for the remote object.
        '''
        classes = self.sharinfo.get('classes')
        return classes

    async def getPoolLink(self):

        while self.links and not self.isfini:

            link = self.links.popleft()

            if link.isfini:
                continue

            return link

        # we need a new one...
        return await self._initPoolLink()

    async def getPipeline(self, genr, name=None):
        '''
        Construct a proxy API call pipeline in order to make
        multiple telepath API calls while minimizing round trips.

        Args:
            genr (async generator): An async generator that yields todo tuples.
            name (str): The name of the shared object on the daemon.

        Example:

            def genr():
                yield s_common.todo('getFooByBar', 10)
                yield s_common.todo('getFooByBar', 20)

            for retn in proxy.getPipeline(genr()):
                valu = s_common.result(retn)
        '''
        async with await Pipeline.anit(self, genr, name=name) as pipe:
            async for retn in pipe:
                yield retn

    async def _initPoolLink(self):

        # TODO loop / backoff

        if self.link.get('unix'):

            path = self.link.get('path')
            link = await s_link.unixconnect(path)

        else:

            ssl = self.link.get('ssl')
            host = self.link.get('host')
            port = self.link.get('port')

            link = await s_link.connect(host, port, ssl=ssl)

        self.onfini(link)

        return link

    async def _putPoolLink(self, link):

        if link.isfini:
            return

        # If we've exceeded our poolsize, discard the current link.
        if len(self.links) >= self._link_poolsize:
            return await link.fini()

        self.links.append(link)

    def __enter__(self):
        '''
        Convenience function to enable using Proxy objects as synchronous context managers.

        Note:
            This must not be used from async code, and it should never be used in core synapse code.
        '''
        if s_threads.iden() == self.tid:
            raise s_exc.SynErr('Use of synchronous context manager in async code')
        self._ctxobj = self.schedCoroSafePend(self.__aenter__())
        return self

    def __exit__(self, *args):
        '''
        Note:
            This should never be used by core synapse code.
        '''
        return self.schedCoroSafePend(self._ctxobj.__aexit__(*args))

    async def _onShareFini(self, mesg):

        iden = mesg[1].get('share')
        share = self.shares.get(iden)
        if share is None:
            return

        share.txfini = False
        await share.fini()

    async def _onShareData(self, mesg):

        data = mesg[1].get('data')
        iden = mesg[1].get('share')

        share = self.shares.get(iden)
        if share is None:
            return

        await share._onShareData(data)

    async def call(self, methname, *args, **kwargs):
        '''
        Call a remote method by name.

        Args:
            methname (str): The name of the remote method.
            *args: Arguments to the method call.
            **kwargs: Keyword arguments to the method call.

        Most use cases will likely use the proxy methods directly:

        The following two are effectively the same:

            valu = proxy.getFooBar(x, y)
            valu = proxy.call('getFooBar', x, y)
        '''
        todo = (methname, args, kwargs)
        return await self.task(todo)

    async def taskv2(self, todo, name=None):

        mesg = ('t2:init', {
                'todo': todo,
                'name': name,
                'sess': self.sess})

        link = await self.getPoolLink()

        await link.tx(mesg)

        mesg = await link.rx()
        if mesg is None:
            raise s_exc.LinkShutDown(mesg='Remote peer disconnected')

        if mesg[0] == 't2:fini':
            await self._putPoolLink(link)
            retn = mesg[1].get('retn')
            return s_common.result(retn)

        if mesg[0] == 't2:genr':

            async def genrloop():

                try:

                    while True:

                        mesg = await link.rx()
                        if mesg is None:
                            raise s_exc.LinkShutDown(mesg=mesg)

                        if mesg[0] != 't2:yield':  # pragma: no cover
                            info = 'Telepath protocol violation:  unexpected message received'
                            raise s_exc.BadMesgFormat(mesg=info)

                        retn = mesg[1].get('retn')
                        if retn is None:
                            await self._putPoolLink(link)
                            return

                        # if this is an exception, it's the end...
                        if not retn[0]:
                            await self._putPoolLink(link)

                        yield s_common.result(retn)

                except GeneratorExit:
                    # if they bail early on the genr, fini the link
                    await link.fini()

            return s_coro.GenrHelp(genrloop())

        if mesg[0] == 't2:share':
            iden = mesg[1].get('iden')
            sharinfo = mesg[1].get('sharinfo')
            await self._putPoolLink(link)
            return await Share.anit(self, iden, sharinfo)

    async def task(self, todo, name=None):

        if self.isfini:
            raise s_exc.IsFini(mesg='Telepath Proxy isfini')

        if self.sess is not None:
            return await self.taskv2(todo, name=name)

        task = Task()

        mesg = ('task:init', {
                'task': task.iden,
                'todo': todo,
                'name': name, })

        self.tasks[task.iden] = task

        try:

            await self.link.tx(mesg)
            retn = await task.result()
            return s_common.result(retn)

        finally:
            self.tasks.pop(task.iden, None)

    async def handshake(self, auth=None):

        mesg = ('tele:syn', {
            'auth': auth,
            'vers': televers,
            'name': self.name,
        })

        await self.link.tx(mesg)

        self.synack = await self.link.rx()
        if self.synack is None:
            mesg = 'socket closed by server before handshake'
            raise s_exc.LinkShutDown(mesg=mesg)

        self.sess = self.synack[1].get('sess')
        self.sharinfo = self.synack[1].get('sharinfo', {})
        self.methinfo = self.sharinfo.get('meths', {})

        vers = self.synack[1].get('vers')
        if vers[0] != televers[0]:
            raise s_exc.BadMesgVers(myver=televers, hisver=vers)

        async def rxloop():

            while not self.link.isfini:

                mesg = await self.link.rx()
                if mesg is None:
                    return

                try:

                    func = self.handlers.get(mesg[0])
                    if func is None:
                        logger.warning('Proxy.rxloop: Invalid Message: %r' % (mesg,))
                        return

                    await func(mesg)

                except asyncio.CancelledError:  # pragma: no cover  TODO:  remove once >= py 3.8 only
                    raise

                except Exception:
                    logger.exception('Proxy.rxloop for %r' % (mesg,))

        retn = self.synack[1].get('retn')
        valu = s_common.result(retn)

        self.schedCoro(rxloop())

        return valu

    async def _txShareExc(self, iden):
        # send a share:fini for an unhandled share.
        await self.link.tx(
            ('share:fini', {'share': iden, 'isexc': True})
        )

    async def _onTaskFini(self, mesg):

        # handle task:fini message
        iden = mesg[1].get('task')

        task = self.tasks.pop(iden, None)
        if task is None:
            logger.warning('task:fini for invalid task: %r' % (iden,))
            return

        retn = mesg[1].get('retn')
        type = mesg[1].get('type')

        if type is None:
            return task.reply(retn)

        ctor = sharetypes.get(type, Share)
        item = await ctor.anit(self, retn[1])

        return task.reply((True, item))

    def __getattr__(self, name):

        info = self.methinfo.get(name)
        if info is not None and info.get('genr'):
            meth = GenrMethod(self, name)
            setattr(self, name, meth)
            return meth

        meth = Method(self, name)
        setattr(self, name, meth)
        return meth

class Client(s_base.Base):
    '''
    A Telepath client object which reconnects and allows waiting for link up.

    Notes:

        The conf data allows changing parameters such as timeouts, retry period, and link pool size. The default
        conf data can be seen below::

            conf = {
                'timeout': 10,
                'retrysleep': 0.2,
                'link_poolsize': 4,
            }

    '''
    async def __anit__(self, url, opts=None, conf=None, onlink=None):

        await s_base.Base.__anit__(self)

        if conf is None:
            conf = {}

        if opts is None:
            opts = {}

        self._t_url = url
        self._t_urls = collections.deque()

        self._t_opts = opts
        self._t_conf = conf

        self._t_proxy = None
        self._t_ready = asyncio.Event()
        self._t_onlinks = []
        self._t_methinfo = None
        self._t_named_meths = set()

        if onlink is not None:
            self._t_onlinks.append(onlink)

        async def fini():
            if self._t_proxy is not None:
                await self._t_proxy.fini()
            # Wake any waiters which may be waiting on waitready() calls so those
            # without timeouts specified are not waiting forever.
            self._t_ready.set()

        self.onfini(fini)

        await self._fireLinkLoop()

    def _initUrlDeque(self):
        self._t_urls.clear()
        if isinstance(self._t_url, str):
            self._t_urls.append(self._t_url)
            return
        self._t_urls.extend(self._t_url)

    def _getNextUrl(self):
        # TODO url list in deque
        if not self._t_urls:
            self._initUrlDeque()
        return self._t_urls.popleft()

    def _setNextUrl(self, url):
        self._t_urls.appendleft(url)

    async def onlink(self, func):
        self._t_onlinks.append(func)
        if self._t_proxy:
            await func(self._t_proxy)

    async def offlink(self, func):
        self._t_onlinks.remove(func)

    async def _fireLinkLoop(self):
        self._t_proxy = None
        self._t_ready.clear()
        self.schedCoro(self._teleLinkLoop())

    async def _teleLinkLoop(self):
        lastlog = 0.0

        while not self.isfini:

            url = self._getNextUrl()

            try:
                await self._initTeleLink(url)
                self._t_ready.set()
                return

            except s_exc.TeleRedir as e:
                self._setNextUrl(e.errinfo.get('url'))
                continue

            except asyncio.CancelledError:  # pragma: no cover  TODO:  remove once >= py 3.8 only
                raise

            except Exception as e:
                now = time.monotonic()
                if now > lastlog + 60.0:  # don't logspam the disconnect message more than 1/min
                    logger.exception(f'telepath client ({s_urlhelp.sanitizeUrl(url)}) encountered an error: {e}')
                    lastlog = now
                await self.waitfini(timeout=self._t_conf.get('retrysleep', 0.2))

    async def proxy(self, timeout=10):
        await self.waitready(timeout=timeout)
        ret = self._t_proxy
        if ret is None or ret.isfini is True:
            raise s_exc.IsFini(mesg='Telepath Client Proxy is not available.')
        return ret

    async def _initTeleLink(self, url):
        self._t_proxy = await openurl(url, **self._t_opts)
        self._t_methinfo = self._t_proxy.methinfo
        self._t_proxy._link_poolsize = self._t_conf.get('link_poolsize', 4)

        async def fini():
            if self._t_named_meths:
                for name in self._t_named_meths:
                    delattr(self, name)
                self._t_named_meths.clear()
            if not self.isfini:
                await self._fireLinkLoop()

        self._t_proxy.onfini(fini)

        for onlink in self._t_onlinks:
            try:
                await onlink(self._t_proxy)
                # in case the callback fini()s the proxy
                if self._t_proxy is None:
                    break
            except asyncio.CancelledError:  # pragma: no cover
                raise
            except Exception as e:
                logger.exception(f'onlink: {onlink}')

    async def task(self, todo, name=None):
        # implement the main workhorse method for a proxy to allow Method
        # objects to use us as the proxy.
        while not self.isfini:
            try:
                await self.waitready()

                # there is a small race where the daemon may fini the proxy
                # account for that here...
                if self._t_proxy is None or self._t_proxy.isfini:
                    self._t_ready.clear()
                    continue

                return await self._t_proxy.task(todo, name=name)

            except s_exc.TeleRedir as e:
                url = e.errinfo.get('url')
                self._setNextUrl(url)
                logger.warning(f'telepath task redirected: ({s_urlhelp.sanitizeUrl(url)})')
                await self._t_proxy.fini()

        raise s_exc.IsFini(mesg='Telepath Client isfini')

    async def waitready(self, timeout=10):
        await asyncio.wait_for(self._t_ready.wait(), self._t_conf.get('timeout', timeout))

    def __getattr__(self, name):
        if self._t_methinfo is None:
            raise s_exc.NotReady(mesg='Must call waitready() on Client before first method call')

        info = self._t_methinfo.get(name)
        if info is not None and info.get('genr'):
            meth = GenrMethod(self, name)
            setattr(self, name, meth)
            self._t_named_meths.add(name)
            return meth

        meth = Method(self, name)
        self._t_named_meths.add(name)
        setattr(self, name, meth)
        return meth

    def _getSynVers(self):
        '''
        Helper method to retrieve the remote Synapse version from Client
        for the currently connected Proxy.

        Notes:
            This will return None if the synapse version was not supplied
            during the Telepath handshake.

        Returns:
            tuple: A tuple of major, minor, patch information as integers.
        '''
        return self._t_proxy._getSynVers()

    def _getSynCommit(self):
        '''
        Helper method to retrieve the remote Synapse commit hash from Proxy.

        Notes:
            This will return None if the synapse commit hash was not supplied
            during the Telepath handshake.

        Returns:
            str: A string containing the commit hash. This may be a empty string.
        '''
        return self._t_proxy._getSynCommit()

    def _getClasses(self):
        '''
        Helper method to retrieve the classes that comprise the remote object
        for the currently connected Proxy.

        Notes:
            This will return None if the class version was not supplied
            during the Telepath handshake.

        Returns:
            tuple: A tuple of strings containing the class paths for the remote object.
        '''
        return self._t_proxy._getClasses()

async def disc_consul(info):
    '''
    Support for updating a URL info dictionary which came from a protocol+consul:// URL.

    Notes:
        This updates the info-dictionary in place, placing the ``host`` value into
        an ``original_host`` key, and updating ``host`` and ``port``.

        By default we pull the ``host`` value from the catalog ``Address`` value,
        and the ``port`` from the ``ServicePort`` value.

        The following HTTP parameters are supported:

        - consul: This is the consul host (schema, fqdn and port) to connect to.
        - consul_tag: If set, iterate through the catalog results until a result
          is found which matches the tag value. This is a case sensitive match.
        - consul_tag_address: If set, prefer the ``TaggedAddresses`` from the catalog.
        - consul_service_tag_address: If set, prefer the associated value from the
          ``ServiceTaggedAddresses`` field.
        - consul_nosslverify: If set, disables SSL verification.

    '''
    info.setdefault('original_host', info.get('host'))
    service = info.get('original_host')
    host = info.get('consul')
    tag = info.get('consul_tag')  # iterate through entries until a match for tag is present.
    ctag_addr = info.get('consul_tag_address')  # Prefer a taggedAddress if set
    csvc_tag_addr = info.get('consul_service_tag_address')  # Prefer a serviceTaggedAddress if set
    gkwargs = {'raise_for_status': True}
    if info.get('consul_nosslverify'):
        gkwargs['ssl'] = False

    if ctag_addr and csvc_tag_addr:
        mesg = 'Cannot resolve consul values with both consul_tag_address and consul_service_tag_address'
        raise s_exc.BadUrl(mesg=mesg, consul_tag_address=ctag_addr,
                           consul_service_tag_address=csvc_tag_addr)

    url = f'{host}/v1/catalog/service/{service}'
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(url, **gkwargs) as resp:
                if resp.status == 200:
                    found = await resp.json()
                    for entry in found:

                        # If the consul_tag parameter is passed, we will iterate over
                        # the Consul results until we find the service which matches
                        # our requested tag. Otherwise, we will grab data from the
                        # first record from the service catalog.
                        if tag and tag not in entry.get('ServiceTags'):
                            continue

                        if csvc_tag_addr:
                            # Use the ServiceTaggedAddresses
                            info['host'] = entry['ServiceTaggedAddresses'][csvc_tag_addr]['address']
                            info['port'] = entry['ServiceTaggedAddresses'][csvc_tag_addr]['port']
                        elif ctag_addr:
                            # Use the TaggedAddresses values.
                            info['host'] = entry['TaggedAddresses'][ctag_addr]
                            info['port'] = entry['ServicePort']
                        else:
                            # Use the generic service address/port
                            info['host'] = entry['Address']
                            info['port'] = entry['ServicePort']
                        return

    except asyncio.CancelledError:  # pragma: no cover  TODO:  remove once >= py 3.8 only
        raise
    except Exception as e:
        raise s_exc.BadUrl(mesg=f'Unknown error while resolving service name [{service}] via consul [{str(e)}].',
                           name=service, consul=host) from e
    raise s_exc.BadUrl(mesg=f'Unable to resolve service information from consul for [{service}].',
                       name=service, consul=host, tag=tag)

def alias(name):
    '''
    Resolve a telepath alias via ~/.syn/aliases.yaml

    Args:
        name (str): Name of the alias to resolve.

    Notes:
        An exact match against the aliases will always be returned first.
        If no exact match is found and the name contains a '/' in it, the
        value before the slash is looked up and the remainder of the path
        is joined to any result. This is done to support dynamic Telepath
        share names.

    Returns:
        str: The url string, if present in the alias.  None will be returned
        if there are no matches.
    '''
    path = s_common.getSynPath('aliases.yaml')
    if not os.path.isfile(path):
        return None

    conf = s_common.yamlload(path)

    # Is there an exact match - if so, return it.
    url = conf.get(name)
    if url:
        return url

    # Since telepath supports dynamic shared object access,
    # slice a name at the first '/', look up using that value
    # and then append the second value to it.
    dynname = None
    if '/' in name:
        name, dynname = name.split('/', 1)
    url = conf.get(name)
    if url and dynname:
        url = '/'.join([url, dynname])

    return url

@s_glob.synchelp
async def openurl(url, **opts):
    '''
    Open a URL to a remote telepath object.

    Args:
        url (str): A telepath URL.
        **opts (dict): Telepath connect options.

    Returns:
        (synapse.telepath.Proxy): A telepath proxy object.

    The telepath proxy may then be used for sync or async calls:

        proxy = openurl(url)
        value = proxy.getFooThing()

    ... or ...

        proxy = await openurl(url)
        valu = await proxy.getFooThing()

    ... or ...

        async with await openurl(url) as proxy:
            valu = await proxy.getFooThing()
    '''
    info = chopurl(url, **opts)
    return await openinfo(info)

def chopurl(url, **opts):

    if isinstance(url, str):
        if url.find('://') == -1:
            newurl = alias(url)
            if newurl is None:
                raise s_exc.BadUrl(mesg=f':// not found in [{url}] and no alias found!',
                                   url=url)
            url = newurl

        info = s_urlhelp.chopurl(url)

        # flatten query params into info
        query = info.pop('query', None)
        if query is not None:
            info.update(query)

    elif isinstance(url, dict):
        info = dict(url)

    else:
        mesg = 'telepath.chopurl() requires a str or dict.'
        raise s_exc.BadArg(mesg)

    info.update(opts)

    return info

async def openinfo(info):

    scheme = info.get('scheme')

    if scheme == 'aha':
        return await getAhaProxy(info)

    if '+' in scheme:
        scheme, disc = scheme.split('+', 1)
        # Discovery protocols modify info dict inband?
        if disc == 'consul':
            await disc_consul(info)

        else:
            raise s_exc.BadUrl(mesg=f'Unknown discovery protocol [{disc}].',
                               disc=disc)

    host = info.get('host')
    port = info.get('port')

    auth = None

    user = info.get('user')
    if user is not None:
        passwd = info.get('passwd')
        auth = (user, {'passwd': passwd})

    if scheme == 'cell':
        # cell:///path/to/celldir:share
        # cell://rel/path/to/celldir:share
        path = info.get('path')

        name = info.get('name', '*')

        # support cell://<relpath>/<to>/<cell>
        # by detecting host...
        host = info.get('host')
        if host:
            path = path.strip('/')
            path = os.path.join(host, path)

        if ':' in path:
            path, name = path.split(':')

        full = os.path.join(path, 'sock')
        link = await s_link.unixconnect(full)

    elif scheme == 'unix':
        # unix:///path/to/sock:share
        name = '*'
        path = info.get('path')
        if ':' in path:
            path, name = path.split(':')
        link = await s_link.unixconnect(path)

    else:

        path = info.get('path')
        name = info.get('name', path[1:])

        hostname = None

        sslctx = None
        if scheme == 'ssl':

            certdir = info.get('certdir')
            certname = info.get('certname')
            hostname = info.get('hostname', host)

            if certdir is None:
                certdir = s_certdir.getCertDir()

            # if a TLS connection specifies a user with no password
            # attempt to auto-resolve a user certificate for the given
            # host/network.
            if certname is None and user is not None and passwd is None:
                certname = f'{user}@{hostname}'

            sslctx = certdir.getClientSSLContext(certname=certname)

            # do hostname checking manually to avoid DNS lookups
            # ( to support dynamic IP addresses on services )
            sslctx.check_hostname = False

        link = await s_link.connect(host, port, ssl=sslctx, hostname=hostname)

    prox = await Proxy.anit(link, name)
    prox.onfini(link)

    try:
        await prox.handshake(auth=auth)

    except (asyncio.CancelledError, Exception):
        await prox.fini()
        raise

    return prox



import types
import asyncio
import logging

logger = logging.getLogger(__name__)

import synapse.exc as s_exc
import synapse.common as s_common
import synapse.telepath as s_telepath

import synapse.lib.base as s_base
import synapse.lib.coro as s_coro
import synapse.lib.link as s_link
import synapse.lib.scope as s_scope
import synapse.lib.share as s_share
import synapse.lib.certdir as s_certdir
import synapse.lib.reflect as s_reflect

class Sess(s_base.Base):

    async def __anit__(self):

        await s_base.Base.__anit__(self)

        self.items = {}
        self.iden = s_common.guid()
        self.user = None
        self.conninfo = {}

    def getSessItem(self, name):
        return self.items.get(name)

    def setSessItem(self, name, item):
        self.items[name] = item

    def popSessItem(self, name):
        return self.items.pop(name, None)

    def pack(self):
        ret = {'items': {name: f'{item.__module__}.{item.__class__.__name__}' for name, item in self.items.items()},
               'conninfo': self.conninfo,
               }
        if self.user:
            ret['user'] = {'iden': self.user.iden,
                           'name': self.user.name,
                           }
        return ret

class Genr(s_share.Share):

    typename = 'genr'

    async def _runShareLoop(self):

        try:

            for item in self.item:

                if self.isfini:
                    break

                retn = (True, item)
                mesg = ('share:data', {'share': self.iden, 'data': retn})

                await self.link.tx(mesg)

                # purposely yield for fair scheduling
                await asyncio.sleep(0)

        except Exception as e:

            retn = s_common.retnexc(e)
            mesg = ('share:data', {'share': self.iden, 'data': retn})
            await self.link.tx(mesg)

        finally:

            mesg = ('share:data', {'share': self.iden, 'data': None})
            await self.link.tx(mesg)
            await self.fini()


class AsyncGenr(s_share.Share):

    typename = 'genr'

    async def _runShareLoop(self):

        try:

            async for item in self.item:

                if self.isfini:
                    break

                retn = (True, item)
                mesg = ('share:data', {'share': self.iden, 'data': retn})

                await self.link.tx(mesg)

                # purposely yield for fair scheduling
                await asyncio.sleep(0)

        except Exception as e:
            retn = s_common.retnexc(e)
            mesg = ('share:data', {'share': self.iden, 'data': retn})
            await self.link.tx(mesg)

        finally:

            mesg = ('share:data', {'share': self.iden, 'data': None})
            await self.link.tx(mesg)
            await self.fini()

dmonwrap = (
    (s_coro.GenrHelp, AsyncGenr),
    (types.AsyncGeneratorType, AsyncGenr),
    (types.GeneratorType, Genr),
)

async def t2call(link, meth, args, kwargs):
    '''
    Call the given ``meth(*args, **kwargs)`` and handle the response to provide
    telepath task v2 events to the given link.
    '''
    try:

        valu = meth(*args, **kwargs)

        if s_coro.iscoro(valu):
            valu = await valu

        try:

            first = True
            if isinstance(valu, types.AsyncGeneratorType):

                async for item in valu:

                    if first:
                        await link.tx(('t2:genr', {}))
                        first = False

                    await link.tx(('t2:yield', {'retn': (True, item)}))

                if first:
                    await link.tx(('t2:genr', {}))

                await link.tx(('t2:yield', {'retn': None}))
                return

            elif isinstance(valu, types.GeneratorType):

                for item in valu:

                    if first:
                        await link.tx(('t2:genr', {}))
                        first = False

                    await link.tx(('t2:yield', {'retn': (True, item)}))

                if first:
                    await link.tx(('t2:genr', {}))

                await link.tx(('t2:yield', {'retn': None}))
                return

        except s_exc.DmonSpawn as e:
            context = e.__context__
            if context:
                if not isinstance(context, asyncio.CancelledError):
                    logger.error('Error during DmonSpawn call: %r', context)
                await link.fini()
            return

        except (asyncio.CancelledError, Exception) as e:

            if isinstance(e, asyncio.CancelledError):
                logger.info('t2call task %s cancelled', meth.__name__)
            else:
                logger.exception('error during task %s', meth.__name__)

            if isinstance(valu, types.AsyncGeneratorType):
                await valu.aclose()
            elif isinstance(valu, types.GeneratorType):
                valu.close()

            if not link.isfini:

                if first:
                    await link.tx(('t2:genr', {}))

                retn = s_common.retnexc(e)
                await link.tx(('t2:yield', {'retn': retn}))

            return

        if isinstance(valu, s_share.Share):

            info = s_reflect.getShareInfo(valu)
            await link.tx(('t2:share', {'iden': valu.iden, 'sharinfo': info}))
            return valu

        await link.tx(('t2:fini', {'retn': (True, valu)}))

    except s_exc.DmonSpawn as e:
        context = e.__context__
        if context:
            logger.error('Error during DmonSpawn call: %r', context)
            await link.fini()
        return

    except (asyncio.CancelledError, Exception) as e:
        logger.exception('error during task: %s', meth.__name__)
        if not link.isfini:
            retn = s_common.retnexc(e)
            await link.tx(('t2:fini', {'retn': retn}))

class Daemon(s_base.Base):

    async def __anit__(self, certdir=None):

        await s_base.Base.__anit__(self)

        self._shareLoopTasks = set()

        if certdir is None:
            certdir = s_certdir.getCertDir()

        self.certdir = certdir
        self.televers = s_telepath.televers

        self.addr = None    # our main listen address
        self.cells = {}     # all cells are shared.  not all shared are cells.
        self.shared = {}    # objects provided by daemon
        self.listenservers = []  # the sockets we're listening on
        self.links = set()

        self.sessions = {}

        self.mesgfuncs = {
            'tele:syn': self._onTeleSyn,
            'task:init': self._onTaskInit,
            'share:fini': self._onShareFini,

            # task version 2 API
            't2:init': self._onTaskV2Init,
        }

        self.onfini(self._onDmonFini)

        # by default we are ready... ( backward compat )
        self.dmonready = True

    async def setReady(self, ready):
        self.dmonready = ready
        if not self.dmonready:
            for link in list(self.links):
                await link.fini()

    async def listen(self, url, **opts):
        '''
        Bind and listen on the given host/port with possible SSL.

        Args:
            host (str): A hostname or IP address.
            port (int): The TCP port to bind.
        '''
        info = s_telepath.chopurl(url, **opts)
        info.update(opts)

        scheme = info.get('scheme')

        if scheme == 'unix':
            path = info.get('path')
            try:
                server = await s_link.unixlisten(path, self._onLinkInit)
            except Exception as e:
                if 'path too long' in str(e):
                    logger.error('unix:// exceeds OS supported UNIX socket path length: %s', path)
                raise

        else:

            host = info.get('host')
            port = info.get('port')

            sslctx = None
            if scheme == 'ssl':

                caname = info.get('ca')
                hostname = info.get('hostname', host)

                sslctx = self.certdir.getServerSSLContext(hostname=hostname, caname=caname)

            server = await s_link.listen(host, port, self._onLinkInit, ssl=sslctx)

        self.listenservers.append(server)
        ret = server.sockets[0].getsockname()

        if self.addr is None:
            self.addr = ret

        return ret

    def share(self, name, item):
        '''
        Share an object via the telepath protocol.

        Args:
            name (str): Name of the shared object
            item (object): The object to share over telepath.
        '''

        try:

            if isinstance(item, s_telepath.Aware):
                item.onTeleShare(self, name)

            self.shared[name] = item

        except Exception:
            logger.exception('onTeleShare() error for: %s)', name)

    async def getSessInfo(self):
        return [sess.pack() for sess in self.sessions.values()]

    async def _onDmonFini(self):
        for s in self.listenservers:
            try:
                s.close()
            except Exception as e:  # pragma: no cover
                logger.warning('Error during socket server close()', exc_info=e)

        finis = [sess.fini() for sess in list(self.sessions.values())]
        if finis:
            await asyncio.wait(finis)

        finis = [link.fini() for link in self.links]
        if finis:
            await asyncio.wait(finis)

        for _, share in self.shared.items():
            if isinstance(share, s_base.Base):
                await share.fini()

    async def _onLinkInit(self, link):

        if not self.dmonready:
            logger.warning(f'onLinkInit is not ready: {repr(link)}')
            return await link.fini()

        self.links.add(link)

        async def fini():
            self.links.discard(link)

        link.onfini(fini)

        async def rxloop():

            task = None
            while not link.isfini:

                mesg = await link.rx()
                if mesg is None:
                    await link.fini()
                    return

                if task is not None:
                    await task

                coro = self._onLinkMesg(link, mesg)
                task = link.schedCoro(coro)

        link.schedCoro(rxloop())

    async def _onLinkMesg(self, link, mesg):

        try:
            func = self.mesgfuncs.get(mesg[0])
            if func is None:
                logger.exception('Dmon.onLinkMesg Invalid: %.80r' % (mesg,))
                return

            await func(link, mesg)

        except asyncio.CancelledError:  # pragma: no cover  # TODO: remove once >= py 3.8 only
            raise

        except ConnectionResetError:
            logger.debug('Dmon.onLinkMesg Handler: connection reset')

        except Exception:
            logger.exception('Dmon.onLinkMesg Handler: %.80r' % (mesg,))

    async def _onShareFini(self, link, mesg):

        sess = link.get('sess')
        if sess is None:
            return

        name = mesg[1].get('share')

        item = sess.popSessItem(name)
        if item is None:
            return

        await item.fini()

    async def _onTeleSyn(self, link, mesg):

        reply = ('tele:syn', {
            'vers': self.televers,
            'retn': (True, None),
        })

        try:

            vers = mesg[1].get('vers')

            if vers[0] != s_telepath.televers[0]:
                raise s_exc.BadMesgVers(vers=vers, myvers=s_telepath.televers)

            path = ()

            name = mesg[1].get('name')
            if not name:
                name = '*'

            if '/' in name:
                name, rest = name.split('/', 1)
                if rest:
                    path = rest.split('/')

            item = self.shared.get(name)

            if item is None:
                raise s_exc.NoSuchName(name=name)

            sess = await Sess.anit()

            async def sessfini():
                self.sessions.pop(sess.iden, None)

            sess.onfini(sessfini)
            link.onfini(sess.fini)

            self.sessions[sess.iden] = sess
            sess.conninfo = link.getAddrInfo()

            link.set('sess', sess)

            if isinstance(item, s_telepath.Aware):
                item = await s_coro.ornot(item.getTeleApi, link, mesg, path)
                if isinstance(item, s_base.Base):
                    link.onfini(item)

            reply[1]['sharinfo'] = s_reflect.getShareInfo(item)

            sess.setSessItem(None, item)
            reply[1]['sess'] = sess.iden

        except Exception as e:
            logger.exception('tele:syn error')
            reply[1]['retn'] = s_common.retnexc(e)

        await link.tx(reply)

    async def _runTodoMeth(self, link, meth, args, kwargs):

        valu = meth(*args, **kwargs)

        for wraptype, wrapctor in dmonwrap:
            if isinstance(valu, wraptype):
                return await wrapctor.anit(link, valu)

        if s_coro.iscoro(valu):
            valu = await valu

        return valu

    def _getTaskFiniMesg(self, task, valu):

        if not isinstance(valu, s_share.Share):
            retn = (True, valu)
            return ('task:fini', {'task': task, 'retn': retn})

        retn = (True, valu.iden)
        typename = valu.typename
        return ('task:fini', {'task': task, 'retn': retn, 'type': typename})

    async def _onTaskV2Init(self, link, mesg):

        # t2:init is used by the pool sockets on the client
        name = mesg[1].get('name')
        sidn = mesg[1].get('sess')
        todo = mesg[1].get('todo')

        try:

            if sidn is None or todo is None:
                raise s_exc.NoSuchObj(name=name)

            sess = self.sessions.get(sidn)
            if sess is None:
                raise s_exc.NoSuchObj(name=name)

            item = sess.getSessItem(name)
            if item is None:
                raise s_exc.NoSuchObj(name=name)

            s_scope.set('sess', sess)
            s_scope.set('link', link)

            methname, args, kwargs = todo

            if methname[0] == '_':
                raise s_exc.NoSuchMeth(name=methname)

            meth = getattr(item, methname, None)
            if meth is None:
                logger.warning('%r has no method: %r', item, methname)
                raise s_exc.NoSuchMeth(name=methname)

            sessitem = await t2call(link, meth, args, kwargs)
            if sessitem is not None:
                sess.onfini(sessitem)

        except (asyncio.CancelledError, Exception) as e:
            logger.exception('on t2:init: %r' % (mesg,))
            if not link.isfini:
                retn = s_common.retnexc(e)
                await link.tx(('t2:fini', {'retn': retn}))

    async def _onTaskInit(self, link, mesg):

        task = mesg[1].get('task')
        name = mesg[1].get('name')

        sess = link.get('sess')
        if sess is None:
            raise s_exc.NoSuchObj(name=name)

        item = sess.getSessItem(name)
        if item is None:
            raise s_exc.NoSuchObj(name=name)

        try:

            methname, args, kwargs = mesg[1].get('todo')

            if methname[0] == '_':
                raise s_exc.NoSuchMeth(name=methname)

            meth = getattr(item, methname, None)
            if meth is None:
                logger.warning('%r has no method: %s', item, methname)
                raise s_exc.NoSuchMeth(name=methname)

            valu = await self._runTodoMeth(link, meth, args, kwargs)

            mesg = self._getTaskFiniMesg(task, valu)

            await link.tx(mesg)

            # if it's a Share(), spin off the share loop
            if isinstance(valu, s_share.Share):

                if isinstance(item, s_base.Base):
                    item.onfini(valu)

                async def spinshareloop():
                    try:
                        await valu._runShareLoop()
                    except asyncio.CancelledError:
                        pass
                    except Exception:
                        logger.exception('Error running %r', valu)
                    finally:
                        await valu.fini()

                self.schedCoro(spinshareloop())

        except (asyncio.CancelledError, Exception) as e:

            logger.exception('on task:init: %r', mesg)

            retn = s_common.retnexc(e)

            await link.tx(
                ('task:fini', {'task': task, 'retn': retn})
            )



import csv
from unittest import mock

import synapse.common as s_common
import synapse.telepath as s_telepath

import synapse.tests.utils as s_t_utils

import synapse.tools.csvtool as s_csvtool

csvfile = b'''ipv4,fqdn,notes
1.2.3.4,vertex.link,malware
8.8.8.8,google.com,whitelist
'''

csvstorm = b'''
    for ($ipv4, $fqdn, $note) in $rows {
        $lib.print("oh hai")
        [ inet:dns:a=($fqdn,$ipv4) ]
    }
'''

csvfile_missing = b'''fqdn,email,tag
vertex.link,,mytag
google.com,myemail@email.com,
yahoo.com,foo@bar.com,mytag
'''

csvstorm_missing = b'''
    for ($fqdn, $email, $tag) in $rows {
        $lib.print("hello hello")
        [ inet:dns:soa=$lib.guid() :fqdn=$fqdn :email?=$email +?#$tag ]
    }
'''

# count is used for test coverage.
csvstorm_export = b'''
test:int $lib.csv.emit($node, $node.props.loc) | count
'''

class CsvToolTest(s_t_utils.SynTest):

    def _getOldSynVers(self):
        return (0, 0, 0)

    async def test_csvtool(self):

        async with self.getTestCore() as core:

            url = core.getLocalUrl()

            dirn = s_common.gendir(core.dirn, 'junk')

            logpath = s_common.genpath(dirn, 'csvtest.log')

            csvpath = s_common.genpath(dirn, 'csvtest.csv')
            with s_common.genfile(csvpath) as fd:
                fd.write(csvfile)

            stormpath = s_common.genpath(dirn, 'csvtest.storm')
            with s_common.genfile(stormpath) as fd:
                fd.write(csvstorm)

            argv = ['--csv-header', '--debug', '--cortex', url, '--logfile', logpath, stormpath, csvpath]
            outp = self.getTestOutp()

            await s_csvtool.main(argv, outp=outp)
            outp.expect('oh hai')
            outp.expect('2 nodes')

            with mock.patch('synapse.telepath.Proxy._getSynVers', self._getOldSynVers):
                outp = self.getTestOutp()
                await s_csvtool.main(argv, outp=outp)
                outp.expect('Cortex version 0.0.0 is outside of the csvtool supported range')

    async def test_csvtool_missingvals(self):

        async with self.getTestCore() as core:

            url = core.getLocalUrl()

            dirn = s_common.gendir(core.dirn, 'junk')

            logpath = s_common.genpath(dirn, 'csvtest.log')

            csvpath = s_common.genpath(dirn, 'csvtest.csv')
            with s_common.genfile(csvpath) as fd:
                fd.write(csvfile_missing)

            stormpath = s_common.genpath(dirn, 'csvtest.storm')
            with s_common.genfile(stormpath) as fd:
                fd.write(csvstorm_missing)

            argv = ['--csv-header', '--debug', '--cortex', url, '--logfile', logpath, stormpath, csvpath]
            outp = self.getTestOutp()

            await s_csvtool.main(argv, outp=outp)
            outp.expect('hello hello')
            outp.expect("'fqdn': 'google.com'")
            outp.expect('3 nodes')

    async def test_csvtool_local(self):

        with self.getTestDir() as dirn:

            logpath = s_common.genpath(dirn, 'csvtest.log')

            csvpath = s_common.genpath(dirn, 'csvtest.csv')
            with s_common.genfile(csvpath) as fd:
                fd.write(csvfile)

            stormpath = s_common.genpath(dirn, 'csvtest.storm')
            with s_common.genfile(stormpath) as fd:
                fd.write(csvstorm)

            argv = ['--csv-header', '--debug', '--test', '--logfile', logpath, stormpath, csvpath]
            outp = self.getTestOutp()

            await s_csvtool.main(argv, outp=outp)
            outp.expect('2 nodes')

    async def test_csvtool_cli(self):

        with self.getTestDir() as dirn:

            logpath = s_common.genpath(dirn, 'csvtest.log')

            csvpath = s_common.genpath(dirn, 'csvtest.csv')
            with s_common.genfile(csvpath) as fd:
                fd.write(csvfile)

            stormpath = s_common.genpath(dirn, 'csvtest.storm')
            with s_common.genfile(stormpath) as fd:
                fd.write(csvstorm)

            argv = ['--csv-header', '--debug', '--cli', '--test', '--logfile', logpath, stormpath, csvpath]
            outp = self.getTestOutp()

            cmdg = s_t_utils.CmdGenerator(['storm --hide-props inet:fqdn',
                                           EOFError(),
                                           ])

            with self.withCliPromptMockExtendOutp(outp):
                with self.withTestCmdr(cmdg):
                    await s_csvtool.main(argv, outp=outp)

            outp.expect('inet:fqdn=google.com')
            outp.expect('2 nodes')

    async def test_csvtool_export(self):

        async with self.getTestCore() as core:

            await core.nodes('[ test:int=20 :loc=us ]')
            await core.nodes('[ test:int=30 :loc=cn ]')
            await core.nodes('[ test:int=40 ]')

            url = core.getLocalUrl()

            dirn = s_common.gendir(core.dirn, 'junk')

            csvpath = s_common.genpath(dirn, 'csvtest.csv')

            stormpath = s_common.genpath(dirn, 'csvtest.storm')
            with s_common.genfile(stormpath) as fd:
                fd.write(csvstorm_export)

            # test a few no-no cases
            argv = ['--test', '--export', stormpath, csvpath]
            outp = self.getTestOutp()
            await s_csvtool.main(argv, outp=outp)
            outp.expect('--export requires --cortex')

            argv = ['--cortex', url, '--export', stormpath, csvpath, 'lol.csv']
            outp = self.getTestOutp()
            await s_csvtool.main(argv, outp=outp)
            outp.expect('--export requires exactly 1 csvfile')

            argv = ['--cortex', url, '--export', stormpath, csvpath]
            outp = self.getTestOutp()

            await s_csvtool.main(argv, outp=outp)

            outp.expect('Counted 3 nodes.')
            outp.expect('3 csv rows')

            with open(csvpath, 'r') as fd:
                rows = [row for row in csv.reader(fd)]
                self.eq(rows, (['20', 'us'], ['30', 'cn'], ['40', '']))

            with mock.patch('synapse.telepath.Proxy._getSynVers', self._getOldSynVers):
                outp = self.getTestOutp()
                await s_csvtool.main(argv, outp=outp)
                outp.expect(f'Cortex version 0.0.0 is outside of the csvtool supported range')



import synapse.exc as s_exc
import synapse.tests.utils as s_t_utils

class GeoPolModelTest(s_t_utils.SynTest):

    async def test_forms_country(self):
        async with self.getTestCore() as core:
            formname = 'pol:country'
            guid = 32 * '0'
            flag_valu = 'sha256:' + 64 * 'f'

            input_props = {'flag': flag_valu, 'founded': 456, 'iso2': 'VI', 'iso3': 'VIS', 'isonum': 31337,
                           'name': 'Republic of Visi', 'tld': 'visi', 'pop': 123}
            expected_props = {'flag': flag_valu, 'founded': 456, 'iso2': 'vi', 'iso3': 'vis', 'isonum': 31337,
                              'name': 'republic of visi', 'tld': 'visi', 'pop': 123}
            expected_ndef = (formname, guid)

            async with await core.snap() as snap:
                node = await snap.addNode(formname, guid, props=input_props)

            self.eq(node.ndef, expected_ndef)
            for prop, valu in expected_props.items():
                self.eq(node.get(prop), valu)

    async def test_types_iso2(self):
        async with self.getTestCore() as core:
            t = core.model.type('pol:iso2')

            self.eq(t.norm('Fo'), ('fo', {}))
            self.raises(s_exc.BadTypeValu, t.norm, 'A')
            self.raises(s_exc.BadTypeValu, t.norm, 'asD')

    async def test_types_iso3(self):
        async with self.getTestCore() as core:
            t = core.model.type('pol:iso3')

            self.eq(t.norm('Foo'), ('foo', {}))
            self.raises(s_exc.BadTypeValu, t.norm, 'As')
            self.raises(s_exc.BadTypeValu, t.norm, 'asdF')

    async def test_types_unextended(self):
        # The following types are subtypes that do not extend their base type
        async with self.getTestCore() as core:
            self.nn(core.model.type('pol:country'))  # guid
            self.nn(core.model.type('pol:isonum'))  # int
